(in-package :cl-astar)


(define-let+-expansion (&stack (x) :once-only? nil)
  `(let ((,x ,let-plus::value))
     (declare (dynamic-extent ,x))
     ,@let-plus::body))

(defun lambda-expr-p (expr)
  (or (symbolp expr)
      (and (listp expr)
           (eq (first expr) 'lambda))))

(define-let+-expansion (&wrapper (name fn args ret) :uses-value? nil)
  (let ((arg-names (mapcar #'first args)))
    `(flet ((,name ,arg-names (funcall ,(if (lambda-expr-p fn)
                                            `(function ,fn)
                                            fn)
                                       ,@arg-names)))
       (declare (inline ,name)
                (ftype (function ,(mapcar #'second args) ,ret) ,name))
       ,@let-plus::body)))

;; 

(eval-when (:compile-toplevel :load-toplevel :execute)
  (declaim (type array-length +bits-per-coordinate+))
  (defconstant +bits-per-coordinate+ (truncate (log most-positive-fixnum 2) 2)
    "Number of bits in `INTEGER-COORDINATE`."))

(deftype integer-coordinate ()
  "An integer coordinate type suitable for use with `ENCODE-INTEGER-COORDINATES`
and `DECODE-INTEGER-COORDINATES`."
  `(signed-byte ,+bits-per-coordinate+))

(declaim (inline encode-integer-coordinates)
         (ftype (function (integer-coordinate integer-coordinate) fixnum)
                encode-integer-coordinates))
(defun encode-integer-coordinates (x y)
  "An implemetation of encoder that fits given integer coordinates into single
`FIXNUM` using dirty bit tricks.

See `INTEGER-COORDINATE`"
  (logior (logand x #.(1- (ash 1 +bits-per-coordinate+)))
          (ash y +bits-per-coordinate+)))

(declaim (inline decode-integer-coordinates)
         (ftype (function (fixnum)
                          (values integer-coordinate integer-coordinate))
                decode-integer-coordinates))
(defun decode-integer-coordinates (value)
  "An implementation of decoder that unpacks integer coordinates from single
`FIXNUM` using dirty bit tricks.

See `INTEGER-COORDINATE`"
  (let ((x (logand value #.(1- (ash 1 +bits-per-coordinate+)))))
    (values
     (if (logtest x #.(ash 1 (1- +bits-per-coordinate+)))
         (- x #.(ash 1 +bits-per-coordinate+)) x)
     (ash value #.(- +bits-per-coordinate+)))))

;; 

(eval-when (:compile-toplevel :load-toplevel :execute)
  (declaim (type array-length
                 +bits-per-float-significand+ +bits-per-float-exponent+))
  (defconstant +bits-per-float-significand+ 24)
  (defconstant +bits-per-float-exponent+ (- +bits-per-coordinate+
                                            1 +bits-per-float-significand+))
  (declaim (type single-float +most-float-coordinate+ +least-float-coordinate+))
  (defconstant +most-float-coordinate+
    (scale-float
     (float (1- (ash 1 +bits-per-float-significand+)))
     (1- (ash 1 (1- +bits-per-float-exponent+))))
    "A maximum value `FLOAT-COORDINATE` can take. Values greater than this would
lead to overflows.")
  (defconstant +least-float-coordinate+
    (scale-float
     (float (ash (1- (ash 1 +bits-per-float-significand+)) -2))
     (- (1- (ash 1 (1- +bits-per-float-exponent+)))))
    "A minimum value `FLOAT-COORDINATE` can take. Values smaller than this
(except zero) would lead to underflows."))

(deftype float-coordinate ()
  "A floating-point coordinate type suitable for use with
`ENCODE-FLOAT-COORDINATES` and `DECODE-FLOAT-COORDINATES`."
  `(or (single-float ,(- +most-float-coordinate+)
                     ,(- +least-float-coordinate+))
       (single-float 0.0 0.0)
       (single-float ,+least-float-coordinate+
                     ,+most-float-coordinate+)))

(declaim (inline encode-float-coordinates)
         (ftype (function (float-coordinate float-coordinate) fixnum)
                encode-float-coordinates))
(defun encode-float-coordinates (x y)
  "An implemetation of encoder that fits given floating-point coordinates into
single `FIXNUM` using dirty bit tricks.

See `FLOAT-COORDINATE`"
  (flet ((float->int (f)
           (declare (type single-float f))
           (let+ (((&values significand exponent sign)
                   (integer-decode-float f))
                  (result 0))
             (declare (type (unsigned-byte #.+bits-per-float-significand+)
                            significand)
                      (type (signed-byte #.+bits-per-float-exponent+)
                            exponent)
                      (type (signed-byte 2) sign)
                      (type integer-coordinate result))
             (setf (ldb (byte +bits-per-float-significand+ 0) result)
                   significand
                   (ldb (byte +bits-per-float-exponent+
                              +bits-per-float-significand+)
                        result)
                   (logand exponent #.(1- (ash 1 +bits-per-float-exponent+))))
             (* sign result))))
    (declare (inline float->int))
    (let ((x* (float->int x))
          (y* (float->int y)))
      (declare (type integer-coordinate x* y*))
      (encode-integer-coordinates x* y*))))

(declaim (inline decode-float-coordinates)
         (ftype (function (fixnum) (values single-float single-float))
                decode-float-coordinates))
(defun decode-float-coordinates (value)
  "An implementation of decoder that unpacks floating-point coordinates from
single `FIXNUM` using dirty bit tricks.

See `FLOAT-COORDINATE`"
  (let+ (((&flet int->float (i)
            (declare (type fixnum i))
            (let* ((v (abs i))
                   (significand (ldb (byte +bits-per-float-significand+ 0) v))
                   (exponent (ldb (byte +bits-per-float-exponent+
                                        +bits-per-float-significand+)
                                  v)))
              (when (>= exponent #.(ash 1 (1- +bits-per-float-exponent+)))
                (setf exponent
                      (- exponent #.(ash 1 +bits-per-float-exponent+))))
              (* (signum i) (scale-float (float significand) exponent)))))
         ((&values x* y*) (decode-integer-coordinates value))
         (x (int->float x*))
         (y (int->float y*)))
    (values x y)))

;; 

(defmacro make-row-major-indexer (width &key (node-width 1) (node-height 1))
  "Constructs indexer function suitable for use as `INDEXER` argument to
`DEFINE-PATH-FINDER`. The resulting function returns index according to world
`WIDTH` (in nodes) and `NODE-WIDTH` and `NODE-HEIGHT` dimensions (both 1 by
default)."
  `(lambda (x y)
     (+ (the array-index
             (* (floor y ,node-height) ,width))
        (floor x ,node-width))))

(defmacro make-column-major-indexer (height &key (node-width 1) (node-height 1))
  "Constructs indexer function suitable for use as `INDEXER` argument to
`DEFINE-PATH-FINDER`. The resulting function returns index according to world
`WIDTH` (in nodes) and `NODE-WIDTH` and `NODE-HEIGHT` dimensions (both 1 by
default)."
  `(lambda (x y)
     (+ (the array-index
             (* (floor x ,node-width) ,height))
        (floor y ,node-height))))

;; 

(declaim (inline goal-reached-exact-p))
(defun goal-reached-exact-p (x y goal-x goal-y)
  "An implementation of goal reached predicate that checks whether given
position is identical to the goal postion.

**NOTE**: this is most likely to work when goal coordinate is multiple of
  node (e.g. map tile) size."
  (and (= x goal-x)
       (= y goal-y)))

;; 

(defvar *coordinate-type*)

(defmacro make-4-directions-enumerator (&key (node-width 1) (node-height 1)
                                             (min-x 0) (min-y 0) max-x max-y)
  "Constructs neighbour enumerator function suitable for use as
`NEIGHBOUR-ENUMERATOR` argument to `DEFINE-PATH-FINDER` for the uniform node
grid allowing movement in 4 directions (left, right, down, up).  `NODE-WIDTH`
and `NODE-HEIGHT` are constant values or constant expressions specifying the
node size within the grid, both default to 1. `MIN-X`, `MIN-Y`, `MAX-X` and
`MAX-Y` are values or expressions specifying the constraints for possible node
coordinate values."
  (let ((w  `(coerce ,node-width      ',*coordinate-type*))
        (h  `(coerce ,node-height     ',*coordinate-type*))
        (-w `(coerce (- ,node-width)  ',*coordinate-type*))
        (-h `(coerce (- ,node-height) ',*coordinate-type*))
        (z  `(coerce 0                ',*coordinate-type*)))
    `(lambda (x y f)
       (loop :for dx :of-type ,*coordinate-type*
               :across (load-time-value (vector ,-w ,z ,z ,w) t)
             :for dy :of-type ,*coordinate-type*
               :across (load-time-value (vector ,z ,-h ,h ,z) t)
             :for next-x :of-type ,*coordinate-type* := (+ x dx)
             :for next-y :of-type ,*coordinate-type* := (+ y dy)
             :when (and (>= next-x ,min-x)
                        (>= next-y ,min-y)
                        (< next-x ,max-x)
                        (< next-y ,max-y))
               :do (funcall f next-x next-y)))))

(defmacro make-8-directions-enumerator (&key (node-width 1) (node-height 1)
                                             (min-x 0) (min-y 0) max-x max-y)
  "Constructs neighbour enumerator function suitable for use as
`NEIGHBOUR-ENUMERATOR` argument to `DEFINE-PATH-FINDER` for the uniform node
grid allowing movement in 8 directions (two vertical, two horizontal and four
diagonal).  `NODE-WIDTH` and `NODE-HEIGHT` are constant values or constant
expressions specifying the node size within the grid, both default to
1. `MIN-X`, `MIN-Y`, `MAX-X` and `MAX-Y` are values or expressions specifying
the constraints for possible node coordinate values."
  (let ((w  `(coerce ,node-width      ',*coordinate-type*))
        (h  `(coerce ,node-height     ',*coordinate-type*))
        (-w `(coerce (- ,node-width)  ',*coordinate-type*))
        (-h `(coerce (- ,node-height) ',*coordinate-type*))
        (z  `(coerce 0                ',*coordinate-type*)))
    `(lambda (x y f)
       (loop
         :for dx :of-type ,*coordinate-type*
           :across (load-time-value (vector ,w ,w ,z ,-w ,-w ,-w ,z ,w) t)
         :for dy :of-type ,*coordinate-type*
           :across (load-time-value (vector ,z ,-h ,-h ,-h ,z ,h ,h ,h) t)
         :for next-x :of-type ,*coordinate-type* := (+ x dx)
         :for next-y :of-type ,*coordinate-type* := (+ y dy)
         :when (and (>= next-x ,min-x)
                    (>= next-y ,min-y)
                    (< next-x ,max-x)
                    (< next-y ,max-y))
           :do (funcall f next-x next-y)))))

;; 

(defmacro make-manhattan-distance-heuristic (&optional (scale-factor 1.0))
  "Constructs heuristic cost function suitable for use as `HEURISTIC-COST`
argument to `DEFINE-PATH-FINDER` based on Manhattan distance between two
nodes. `SCALE-FACTOR` is constant expression to multiply the calculated
distance by. Note that this kind of heuristic is best suited for grids
allowing 4 directions of movement.

See `MAKE-4-DIRECTIONS-ENUMERATOR`"
  `(lambda (x1 y1 x2 y2)
     (* ,scale-factor
        (+ (abs (float (- x1 x2)))
           (abs (float (- y1 y2)))))))

(defmacro make-octile-distance-heuristic (&optional (scale-factor 1.0))
  "Constructs heuristic cost function suitable for use as `HEURISTIC-COST`
argument to `DEFINE-PATH-FINDER` based on octile distance between two
nodes. `SCALE-FACTOR` is constant expression to multiply the calculated
distance by. Note that this kind of heuristic is best suited for grids
allowing 8 directions of movement.

See `MAKE-8-DIRECTIONS-ENUMERATOR`"
  `(lambda (x1 y1 x2 y2)
     (let ((dx (abs (float (- x1 x2))))
           (dy (abs (float (- y1 y2)))))
       (* ,scale-factor
          (+ (* ,(sqrt 2) (min dx dy))
             (abs (- dx dy)))))))

(defmacro make-euclidean-distance-heuristic (&optional (scale-factor 1.0))
  "Constructs heuristic cost function suitable for use as `HEURISTIC-COST`
argument to `DEFINE-PATH-FINDER` based on Euclidean distance between two
nodes. `SCALE-FACTOR` is constant expression to multiply the calculated
distance by. Note that this kind of heuristic is best suited for grids
allowing any direction of movement."
  `(lambda (x1 y1 x2 y2)
     (let ((dx (abs (float (- x1 x2))))
           (dy (abs (float (- y1 y2)))))
       (* ,scale-factor
          (sqrt (+ (* dx dx)
                   (* dy dy)))))))

;; 

(defmacro %required-argument (argument)
  `(unless ,argument
     (error ,(format nil "Missing required argument ~a" argument))))

(defun %parse-body (body)
  (let ((doc nil)
        (decls nil)
        (current nil))
    (tagbody
     :declarations
       (setf current (car body))
       (when (stringp current)
         (if doc
             (error "Too many documentation strings in ~S." body)
             (setf doc (pop body)))
         (go :declarations))
       (when (and (listp current) (eq (first current) 'declare))
         (push (pop body) decls)
         (go :declarations)))
    (values body (nreverse decls) doc)))

(defmacro define-path-finder (name (&rest parameters)
                              (&key world-size
                                 (frontier-size 500)
                                 (coordinate-type 'float-coordinate)
                                 (coordinate-encoder 'encode-float-coordinates)
                                 (coordinate-decoder 'decode-float-coordinates)
                                 indexer
                                 goal-reached-p
                                 neighbour-enumerator
                                 exact-cost
                                 heuristic-cost
                                 (max-movement-cost 'most-positive-single-float)
                                 (path-initiator 'identity)
                                 path-processor
                                 (path-finalizer '(constantly t))
                                 variables)
                              &body declarations-and-docstring)
  "Defines a function with `NAME` that takes four values of
`COORDINATE-TYPE` (coordinates of start and goal nodes) and arbitrary number of
keyword parameters specified in `PARAMETERS` list. The function would return
the value returned by `PATH-FINALIZER` function (`T` by default, see below) or
`NIL` if a path from start to goal node could not be found. One can also supply
local declarations and docstring for the function being defined using the
`DECLARATIONS-AND-DOCSTRING` argument.

To adjust the resulting function to your needs, you should specify the
following arguments to the macro:

- `WORLD-SIZE`: expression for node count in the world, could be variable;
  calculated only once per function invocation.

- `FRONTIER-SIZE`: initial A* frontier size (see
  [here](https://redblobgames.com/pathfinding/a-star/introduction.html#breadth-first-search)).
  Default value of 500 is more than enough for sane cases and makes it fit
  into single memory page, which might or might not improve performance.

- `COORDINATE-TYPE`: type of coordinate that meets the aforementioned
  assumptions. There are predefined integer and floating-point types
  `INTEGER-COORDINATE` and `FLOAT-COORDINATE` that can be used
  here. Floating-point coordinate type makes sense for use with
  e.g. [liballegro](https://github.com/resttime/cl-liballegro) (where pixels do
  in fact have floating-point coordinates), integer type makes sense for other
  graphic backends and simple cases where the nodes are just stored in an
  array. Defaults to `FLOAT-COORDINATE`.

- `COORDINATE-ENCODER`: expression evaluating to lambda expression, or unquoted
  symbol denoting (preferably inline) function that takes two coordinates of
  `COORDINATE-TYPE` and returns single `FIXNUM` representing those
  coordinates. Defaults to `ENCODE-FLOAT-COORDINATES`.

- `COORDINATE-DECODER`: expression evaluating to lambda expression, or unquoted
  symbol denoting (preferably inline) function that takes single `FIXNUM` value
  and unpacks it by returning two coordinates of `COORDINATE-TYPE` as multiple
  `VALUES`. Defaults to `DECODE-FLOAT-COORDINATES`.

- `INDEXER`: expression evaluating to lambda expression, or unquoted symbol
  denoting (preferably inline) function that takes two coordinates of
  `COORDINATE-TYPE` and returns index in some 1-dimensional array; the index
  must be unique for the node containing those coordinates. The calls to
  `MAKE-ROW-MAJOR-INDEXER` or `MAKE-COLUMN-MAJOR-INDEXER` macros might be
  suitable values of this argument for cases when nodes are uniform in size.

- `GOAL-REACHED-P`: expression evaluating to lambda expression, or unquoted
  symbol denoting (preferably inline) function that takes four values of
  `COORDINATE-TYPE` (coordinates of current and goal nodes) and returns boolean
  indicating whether the goal node is reached by being in that current
  node. `GOAL-REACHED-EXACT-P` could be used here.

- `NEIGHBOUR-ENUMERATOR`: expression evaluating to lambda expression, or
  unquoted symbol denoting (preferably inline) function that takes a current
  coordinate (two values of `COORDINATE-TYPE`) and a function of two arguments
  of `COORDINATE-TYPE`. `NEIGHBOUR-ENUMERATOR` should call given function with
  coordinates of all neighbours of a node occupying given position. The call to
  `MAKE-4-DIRECTIONS-ENUMERATOR` or `MAKE-8-DIRECTIONS-ENUMERATOR` macros might
  be suitable value of this argument for corresponding cases.

- `EXACT-COST`: expression evaluating to lambda expression, or unquoted symbol
  denoting (preferably inline) function that takes four values of
  `COORDINATE-TYPE` (coordinates of two neighbouring nodes) and returns
  non-negative `SINGLE-FLOAT` corresponding to exact cost to move from one
  given neighbouring node to another. Note that it should be at the same scale
  as `HEURISTIC-COST`. See
  [here](http://theory.stanford.edu/~amitp/GameProgramming/MovementCosts.html)
  for a discussion of movement cost functions.

- `HEURISTIC-COST`: expression evaluating to lambda expression, or unquoted
  symbol denoting (preferably inline) function that takes four values of
  `COORDINATE-TYPE` (coordinates of two arbitrary nodes) and returns
  non-negative `SINGLE-FLOAT` corresponding to heuristic cost to move from one
  given arbitrary node to another. Note that it should be at the same scale as
  `EXACT-COST`. Setting this to `(CONSTANTLY 0.0)` turns A* into Dijkstra's
  algorithm. The call to one of `MAKE-MANHATTAN-DISTANCE-HEURISTIC`,
  `MAKE-OCTILE-DISTANCE-HEURISTIC` or `MAKE-EUCLIDEAN-DISTANCE-HEURISTIC`
  macros might be suitable value of this argument. A good survey on different
  heuristics could be found
  [here](http://theory.stanford.edu/~amitp/GameProgramming/Heuristics.html).

- `MAX-MOVEMENT-COST`: when any of two cost functions above return value
  greater or equal to this, it means that the movement is impossible. Defaults
  to `MOST-POSITIVE-SINGLE-FLOAT`.

- `PATH-INITIATOR`: expression evaluating to lambda expression, or unquoted
  symbol denoting (preferably inline) function that is called once the path
  was successfully found. It takes a single argument which is the count of
  nodes along the found path, including start node and goal node. Note that
  when the path is not found, this function isn't called at all. Defaults to
  `IDENTITY`.

- `PATH-PROCESSOR`: expression evaluating to lambda expression, or unquoted
  symbol denoting (preferably inline) function of two arguments of
  `COORDINATE-TYPE` that is successively called with coordinates of every node
  along the path, including start node and goal node. Note that when the path
  is not found, this function isn't called at all.

- `PATH-FINALIZER`: expression evaluating to lambda expression, or unquoted
  symbol denoting (preferably inline) function of no arguments that is called
  once after `PATH-PROCESSOR` has been called with all points in found
  path. The return value of this function becomes the return value of path
  finder function being defined. Note that when the path is not found, this
  function isn't called at all and defined path finder function returns
  `NIL`. Defaults to `(CONSTANTLY T)`.

`VARIABLES` is a list of variable specifiers with optional initializers that
is put after
[`＆AUX`](https://lispworks.com/documentation/HyperSpec/Body/03_dae.htm)
keyword in parameter list of defined function. Those variables are available
to all lambda expressions listed above. Note that those lambda expressions can
also freely refer to variables `START-X`, `START-Y`, `GOAL-X`, `GOAL-Y` passed
to the function being defined, as well as to all keyword parameters specified
in `PARAMETERS`.

In terms of memory consumption, the defined function uses 12×`FRONTIER-SIZE`
bytes of stack memory every time it is called (note that frontier may grow for
big worlds and long paths) and 20×`WORLD-SIZE` bytes of heap memory, but only
the first time, and every time `WORLD-SIZE` grows bigger (the storage arrays
are closed over the function and reused)."
  (%required-argument world-size)
  (%required-argument indexer)
  (%required-argument goal-reached-p)
  (%required-argument neighbour-enumerator)
  (%required-argument exact-cost)
  (%required-argument heuristic-cost)
  (%required-argument path-processor)
  (setf *coordinate-type* coordinate-type)
  (let+ ((start-x (ensure-symbol :start-x))
         (start-y (ensure-symbol :start-y))
         (goal-x (ensure-symbol :goal-x))
         (goal-y (ensure-symbol :goal-y))
         ((&values body declarations docstring)
          (%parse-body declarations-and-docstring))
         ((&with-gensyms world-size* cost-so-far came-from path
                         coordinate-encoder* x y x2 y2 v f l
                         coordinate-decoder* indexer* goal-reached-p*
                         neighbour-enumerator* exact-cost* heuristic-cost*
                         path-initiator* path-processor*))
         ((&once-only frontier-size)))
    (when body
      (error "Excess forms in DECLARATIONS-AND-DOCSTRING argument: ~s" body))
    `(let ((,cost-so-far (make-array 0 :element-type 'single-float))
           (,came-from (make-array 0 :element-type 'fixnum))
           (,path (make-array 0 :element-type 'fixnum)))
       (declare (type (simple-array single-float) ,cost-so-far)
                (type (simple-array fixnum) ,came-from ,path))
       (defun ,name (,start-x ,start-y ,goal-x ,goal-y &key ,@parameters
                     &aux ,@variables)
         (declare (type ,coordinate-type ,start-x ,start-y ,goal-x ,goal-y))
         ,@declarations
         ,docstring
         (let+ ((,world-size* ,world-size)
                ((&wrapper ,coordinate-encoder* ,coordinate-encoder
                           ((,x ,coordinate-type) (,y ,coordinate-type))
                           fixnum))
                ((&wrapper ,coordinate-decoder* ,coordinate-decoder
                           ((,v fixnum))
                           (values ,coordinate-type ,coordinate-type)))
                ((&wrapper ,indexer* ,indexer
                           ((,x ,coordinate-type) (,y ,coordinate-type))
                           array-index))
                ((&wrapper ,goal-reached-p* ,goal-reached-p
                           ((,x ,coordinate-type) (,y ,coordinate-type)
                            (,x2 ,coordinate-type) (,y2 ,coordinate-type))
                           boolean))
                ((&wrapper ,neighbour-enumerator* ,neighbour-enumerator
                           ((,x ,coordinate-type) (,y ,coordinate-type)
                            (,f (function (,coordinate-type ,coordinate-type))))
                           t))
                ((&wrapper ,exact-cost* ,exact-cost
                           ((,x ,coordinate-type) (,y ,coordinate-type)
                            (,x2 ,coordinate-type) (,y2 ,coordinate-type))
                           (single-float 0.0)))
                ((&wrapper ,heuristic-cost* ,heuristic-cost
                           ((,x ,coordinate-type) (,y ,coordinate-type)
                            (,x2 ,coordinate-type) (,y2 ,coordinate-type))
                           (single-float 0.0)))
                ((&wrapper ,path-initiator* ,path-initiator
                           ((,l array-length)) t))
                ((&wrapper ,path-processor* ,path-processor
                           ((,x ,coordinate-type) (,y ,coordinate-type))
                           t))
                ((&macrolet world-ref (array x y)
                   `(aref ,array (,',indexer* ,x ,y))))
                ((&stack frontier-items)
                 (make-array ,frontier-size :element-type 'fixnum))
                ((&stack frontier-priorities)
                 (make-array ,frontier-size :element-type 'single-float))
                (frontier (make-queue frontier-items frontier-priorities)))
           (declare (type array-length ,world-size*))
           (when (< (length ,cost-so-far) ,world-size*)
             (adjust-simple-arrayf ,cost-so-far ,world-size*
                                   :element-type 'single-float))
           (when (< (length ,came-from) ,world-size*)
             (adjust-simple-arrayf ,came-from ,world-size*
                                   :element-type 'fixnum))
           (when (< (length ,path) ,world-size*)
             (adjust-simple-arrayf ,path ,world-size*
                                   :element-type 'fixnum))
           (fill ,cost-so-far single-float-nan)
           (fill ,came-from most-negative-fixnum)
           (queue-insert frontier (,coordinate-encoder* ,start-x ,start-y) 0.0)
           (setf (world-ref ,cost-so-far ,start-x ,start-y) 0.0)
           (let+ (current
                  current-x current-y
                  ((&flet process-neighbour (next-x next-y)
                     (declare (type ,coordinate-type next-x next-y))
                     (let* ((next (,coordinate-encoder* next-x next-y))
                            (new-cost
                              (+ (world-ref ,cost-so-far current-x current-y)
                                 (,exact-cost*
                                  current-x current-y next-x next-y)
                                 (,heuristic-cost*
                                  current-x current-y next-x next-y)))
                            (next-cost (world-ref ,cost-so-far next-x next-y)))
                       (when (and (< new-cost ,max-movement-cost)
                                  (or (float-nan-p next-cost)
                                      (< new-cost next-cost)))
                         (setf (world-ref ,cost-so-far next-x next-y) new-cost
                               (world-ref ,came-from next-x next-y) current)
                         (queue-insert
                          frontier next
                          (+ new-cost (,heuristic-cost*
                                       next-x next-y ,goal-x ,goal-y))))))))
             (declare (type (or fixnum null) current)
                      (type ,coordinate-type current-x current-y)
                      (inline process-neighbour))
             (loop :when (queue-empty-p frontier)
                     :do (setf current nil)
                         (loop-finish)
                   :do (setf current (queue-pop frontier))
                       (multiple-value-setq (current-x current-y)
                         (,coordinate-decoder* current))
                   :when (,goal-reached-p* current-x current-y ,goal-x ,goal-y)
                     :do (loop-finish)
                   :do (,neighbour-enumerator* current-x current-y
                                               #'process-neighbour)
                   :finally (unless current
                              (return nil))
                            (let ((goal (,coordinate-encoder* ,goal-x ,goal-y))
                                  (length 0))
                              (declare (type fixnum goal)
                                       (type array-length length))
                              (loop :for c :of-type fixnum := goal
                                      :then (let+ (((&values x y)
                                                    (,coordinate-decoder* c)))
                                              (world-ref ,came-from x y))
                                    :until (= c most-negative-fixnum)
                                    :do (setf (aref ,path (incf length)) c))
                              (,path-initiator* length)
                              (loop :for i :from length :downto 1
                                    :do (let+ (((&values x y)
                                                (,coordinate-decoder*
                                                 (aref ,path i))))
                                          (,path-processor* x y)))
                              (return (funcall ,path-finalizer))))))))))
