(in-package #:cl-user)

(defpackage cl-astar
  (:documentation "**NOTE: this software is of alpha quality, and the API is
subject to change.**

`cl-astar` is a Common Lisp library providing heavily optimized yet flexible
A* pathfinding algorithm implementation for 2-dimensional spaces.

[A＊](https://en.wikipedia.org/wiki/A%2A_search_algorithm) is the most popular
algorithm choice for pathfinding in video games and other applications,
because it's fairly flexible and can be used in a wide range of
contexts. For more information, refer to
[Amit's A* Pages](https://theory.stanford.edu/~amitp/GameProgramming) which is
amazing resource on this algorithm.

Despite being quite flexible, this library does some assumptions about the path
finding task, namely:

- the world to navigate in could be represented as a graph of discrete *nodes*
  having some positions, some of which are neighbours to other nodes (think
  rectangular map tiles);
- node position within the world is represented by 2D coordinates
  (think pixel `X` & `Y` coordinates);
- this pair of coordinates can fit into single `FIXNUM` (that is, about 60
  bits on most modern CL implementations, so about 30 bits of information, or
  on the order of 10⁹ different states per values of `X` and `Y`);
- there is cost of moving from one node to another, quantified by a
  non-negative floating-point number;
- the cost of moving from given node to its neighbouring node could always be
  exactly determined;
- the cost of moving from one arbitrary node to another could be at least
  heuristically determined (by e.g. measuring distance between them).

A heart of this library is `DEFINE-PATH-FINDER` macro which defines (in terms
of `DEFUN`) the function to find the path given your constraints, goal
reaching conditions and cost functions, and to process found path the way you
need. Hence one might even call this library a *microframework*.

Here's its usage example that finds path in 10×10 maze which is
represented by 2D array with obstacles denoted by star character and empty
cells by space character:

```language-lisp
(ql:quickload :cl-astar)

(defconstant maze (make-array '(10 10)
                              :element-type 'character
                              :initial-contents '(\"  *   *   \"
                                                  \"* * * * * \"
                                                  \"* * * * * \"
                                                  \"* * * * * \"
                                                  \"* * * * * \"
                                                  \"* * * * * \"
                                                  \"* * * * * \"
                                                  \"* * * * * \"
                                                  \"* * * * * \"
                                                  \"*   *   * \")))
(defconstant width  (second (array-dimensions maze)))
(defconstant height (first  (array-dimensions maze)))

(a*:define-path-finder find-path ()
  (:variables ((result (alexandria:copy-array maze)))
   :world-size (* width height)
   :coordinate-type a*:integer-coordinate
   :coordinate-encoder a*:encode-integer-coordinates
   :coordinate-decoder a*:decode-integer-coordinates
   :indexer (a*:make-row-major-indexer width)
   :goal-reached-p a*:goal-reached-exact-p
   :neighbour-enumerator (a*:make-4-directions-enumerator
                          :max-x width :max-y height)
   :exact-cost (lambda (x1 y1 x2 y2)
                 (declare (ignorable x1 y1))
                 (if (eql (aref maze y2 x2) #\\Space)
                     0.0
                     most-positive-single-float))
   :heuristic-cost (a*:make-manhattan-distance-heuristic)
   :path-processor (lambda (x y) (setf (aref result y x) #\\x))
   :path-finalizer (lambda () result)))

(print (find-path 0 0 9 9))
```

The documentation for the symbols exported from the `CL-ASTAR` package can be
found below.")
  (:nicknames #:a*)
  (:use #:cl #:let-plus #:trivial-adjust-simple-array)
  (:import-from #:alexandria #:array-index #:array-length #:ensure-symbol)
  (:import-from #:float-features #:single-float-nan #:float-nan-p)
  (:export
   #:integer-coordinate
   #:encode-integer-coordinates
   #:decode-integer-coordinates
   #:float-coordinate
   #:encode-float-coordinates
   #:decode-float-coordinates
   #:make-row-major-indexer
   #:make-column-major-indexer
   #:goal-reached-exact-p
   #:make-4-directions-enumerator
   #:make-8-directions-enumerator
   #:make-manhattan-distance-heuristic
   #:make-octile-distance-heuristic
   #:make-euclidean-distance-heuristic
   #:define-path-finder))
